"""Test suite for the PyProjectToml class and get_pyproject_path function."""

from __future__ import annotations

import sys
from pathlib import Path
from typing import TypeVar

import pytest

from check_dependencies.lib import Package
from check_dependencies.pyproject_toml import (
    PyProjectToml,
    _get_pyproject_path,
    _nested_item,
)
from tests.conftest import (
    HATCH,
    PEP631,
    POETRY,
    PYPROJECT_EMPTY,
    PYPROJECT_PROVIDES,
    UV_LEGACY,
)

try:
    import tomllib  # type: ignore[unresolved-import]
except ImportError:  # pragma: no cover
    import toml as tomllib


class TestPyProjectToml:
    """Test suite for the PyProjectToml class."""

    def cfg(self, path: Path, include_dev: bool = False) -> PyProjectToml:
        """Create a PyProjectToml instance."""
        return PyProjectToml(
            cfg=tomllib.loads(path.read_text("utf-8")),
            path=path,
            include_dev=include_dev,
        )

    @pytest.mark.parametrize(
        "pyproject, included_dev, add_expect",
        [
            (PEP631, True, []),
            (PEP631, False, []),
            (POETRY, False, []),
            (POETRY, True, ["test_dev_1", "test_dev_2"]),
            (HATCH, False, []),
            (HATCH, True, ["test_dev_1", "test_dev_2"]),
            (UV_LEGACY, False, []),
            (UV_LEGACY, True, ["test_dev_1", "test_dev_2"]),
        ],
    )
    def test_dependencies(
        self,
        pyproject: Path,
        included_dev: bool,
        add_expect: list[str],
    ) -> None:
        """Test get_declared_dependencies function without included development deps."""
        cfg = self.cfg(pyproject, include_dev=included_dev)
        assert set(cfg.dependencies) == {"test_main", "test_1"}.union(add_expect or {})

    def test_unsupported_dependencies(self) -> None:
        """Test unsupported dependencies in pyproject.toml."""
        cfg = self.cfg(PYPROJECT_EMPTY)
        with pytest.raises(ValueError, match="No dependency management found"):
            _ = cfg.dependencies

    def test_provides_empty(self) -> None:
        """Test that provides returns an empty dict when not configured."""
        cfg = self.cfg(PEP631)
        assert cfg.provides == []

    def test_provides(self) -> None:
        """Test that provides returns the correct mapping."""
        cfg = self.cfg(PYPROJECT_PROVIDES)
        assert cfg.provides == [(Package("test_alias_pkg"), "test_1")]

    @pytest.mark.parametrize(
        "raw, expected",
        [
            ("pyjwt", "pyjwt"),
            ("PyJWT", "pyjwt"),
            ("scikit-learn", "scikit_learn"),
            ("scikit_learn", "scikit_learn"),
            ("SciKit-Learn", "scikit_learn"),
            ("Pillow", "pillow"),
        ],
    )
    def test_provides_normalizes_values(self, raw: str, expected: str) -> None:
        """PyProjectToml.provides normalizes package name values."""
        cfg = PyProjectToml(
            cfg={"tool": {"check-dependencies": {"provides": {raw: "some_import"}}}},
            path=Path("dummy"),
            include_dev=False,
        )
        assert cfg.provides == [(Package(expected), "some_import")]

    @pytest.mark.skipif(sys.platform != "win32", reason="Windows-specific test")
    def test_fails_on_different_paths(self) -> None:
        """Test that PyProjectToml raises when initialized with different paths."""
        with pytest.raises(
            ValueError, match=r"Error finding common path for.*C:.test.*D:.test"
        ):
            PyProjectToml.for_paths(["C:/test", "D:/test"])


class TestNestedItem:
    """Test suite for nested item."""

    _T = TypeVar("_T")

    @pytest.mark.parametrize(
        "key, type_, expected",
        [("a.b.c", int, 1), ("a.b.d", int, 2), ("a.b.x", int, 0)],
    )
    def test_nested_item(self, key: str, type_: type[_T], expected: _T) -> None:
        """Test nested item."""
        prj = PyProjectToml(cfg={"a": {"b": {"c": 1, "d": 2}}}, path=Path())
        assert _nested_item(prj.cfg, key, type_) == expected

    def test_raise_wrong_type(self) -> None:
        """Raise wrong type."""
        prj = PyProjectToml(cfg={"a": 1}, path=Path())
        with pytest.raises(TypeError):
            _nested_item(prj.cfg, "a", str)


class TestGetPyProjectPath:
    """Test suite for the get_pyproject_path function."""

    def test_find_pyproject(self, tmp_path: Path) -> None:
        """Test that get_pyproject_path finds the pyproject.toml file."""
        pyproject = tmp_path / "pyproject.toml"
        pyproject.write_text("[tool.check-dependencies]\n", "utf-8")
        assert _get_pyproject_path(tmp_path) == pyproject

    def test_no_pyproject(self) -> None:
        """Test that get_pyproject_path raises without pyproject.toml."""
        with pytest.raises(FileNotFoundError):
            _get_pyproject_path(Path("/"))



"""Main module for check_dependencies."""

from __future__ import annotations

import ast
import logging
from pathlib import Path
from typing import TYPE_CHECKING

from check_dependencies.builtin_module import BUILTINS
from check_dependencies.lib import Dependency, Package, Packages

if TYPE_CHECKING:
    from collections.abc import Collection, Generator, Iterator

    from check_dependencies.app_config import AppConfig

logger = logging.getLogger("check_dependencies")
ERR_MISSING_DEPENDENCY = 2
ERR_EXTRA_DEPENDENCY = 4
ERR_NO_PYPROJECT = 8
ERR_FILE = 16


def yield_wrong_imports(
    file_names: Collection[str],
    app_cfg: AppConfig,
) -> Generator[str, None, int]:
    """Yield output lines of missing/unused imports.

    :param file_names: List of file paths to analyze.
    :param app_cfg: Application configuration containing dependencies and settings.
        If app_cfg.show_all is True, all imports are shown.
    """
    used_deps: set[Package] = set()
    src_fmt = app_cfg.mk_src_formatter()

    allowed_dependencies = {
        *app_cfg.dependencies,  # dependencies from pyproject.toml
        *app_cfg.known_extra,
        *Package.set(BUILTINS),
        *Package.set(app_cfg.known_missing),
    }
    provides = app_cfg.provides

    exit_status = 0

    for src_pth in _project_files(file_names):
        for cause, module, stmt in _missing_imports_iter(
            src_pth, dependencies=allowed_dependencies, provides=provides
        ):
            if cause not in (Dependency.OK, Dependency.FILE_ERROR):
                exit_status |= ERR_MISSING_DEPENDENCY
            if cause != Dependency.FILE_ERROR:
                used_deps |= provides.packages(module)
            yield from src_fmt(src_pth.as_posix(), cause, module, stmt)

    if superfluous_requirements := [
        msg
        for dep in sorted(
            set(app_cfg.dependencies).difference(used_deps, app_cfg.known_extra),
        )
        for msg in app_cfg.unused_fmt(str(dep))
    ]:
        exit_status |= ERR_EXTRA_DEPENDENCY
        if app_cfg.verbose:
            yield ""
            yield "### Dependencies in config file not used in application:"
            yield f"# Config file: {app_cfg.pyproject_file}"
        yield from superfluous_requirements
    return exit_status


def _project_files(file_names: Collection[str]) -> Iterator[Path]:
    """Collect all Python files in a list of files or directories.

    Ensure no file is visited more than once.

    :param file_names: List of file paths or directories to analyze.
    """
    visited = set()
    for p in map(Path, file_names):
        for p_sub in p.rglob("*.py") if p.is_dir() else [p]:
            # resolve to avoid duplicates from symlinks or different relative paths.
            # Symlink can be pointed outside the project directory.
            if (p_sub_resolved := p_sub.resolve()) not in visited:
                visited.add(p_sub_resolved)
                yield p_sub


def _missing_imports_iter(
    file: Path,
    dependencies: Collection[Package],
    provides: Packages,
) -> Iterator[tuple[Dependency, str, ast.stmt]]:
    """Find missing imports in a Python file.

    :param file: Pyton file to analyze
    :param dependencies: Declared dependencies from pyproject file
    :yields: Tuple of status, module name and optional import statement
    """
    try:
        parsed = ast.parse(file.read_bytes(), filename=file.as_posix())
    except (SyntaxError, OSError, PermissionError, FileNotFoundError):
        logger.warning("Could not parse %s", file, exc_info=False)
        yield Dependency.FILE_ERROR, file.as_posix(), ast.Raise(lineno=-1)
        return
    for module, stmt in _imports_iter(parsed.body):
        pkg_ = provides.packages(module)
        status = Dependency.OK if pkg_.intersection(dependencies) else Dependency.NA
        yield status, module, stmt


def _imports_iter(body: list[ast.stmt]) -> Iterator[tuple[str, ast.stmt]]:
    """Yield all import statements from a body of code.

    :param body: List of AST statements to analyze.
    """
    try:
        iter(body)
    except TypeError:
        # not iterable, so return empty
        return

    for x in body:
        if isinstance(x, ast.Import):
            for alias in x.names:
                yield alias.name, x  # yield x, not alias to get lineno
        elif isinstance(x, ast.ImportFrom) and x.level == 0 and x.module:
            # level > 0 means relative import
            yield x.module, x
        elif hasattr(x, "body"):
            for f in x._fields:
                yield from _imports_iter(getattr(x, f))



"""Test the main module."""

from __future__ import annotations

import ast
import time
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest

from check_dependencies.__main__ import main as cli_main
from check_dependencies.app_config import AppConfig
from check_dependencies.lib import Dependency, Package, Packages
from check_dependencies.main import (
    _imports_iter,
    _missing_imports_iter,
    yield_wrong_imports,
)
from tests.conftest import (
    DATA,
    POETRY,
    POETRY_EXTRA,
    PYPROJECT_CFG,
    PYPROJECT_PROVIDES,
    PYPROJECT_UNICODE,
    SRC,
    SRC_MODULE,
    SRC_UNICODE,
)

if TYPE_CHECKING:
    from collections.abc import Sequence


def test__main__(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test the main module.

    This also tests if all dependencies are defined correctly.
    """
    main_module = Path(__file__).parents[1] / "check_dependencies"
    monkeypatch.setattr("sys.argv", ["check_dependencies", main_module.as_posix()])
    with pytest.raises(SystemExit) as excinfo:
        cli_main()
    assert excinfo.value.code == 0


@pytest.mark.parametrize(
    "argv_provides, expected_provides",
    [
        # single flag
        (["--provides", "Pillow=PIL"], [("pillow", {"PIL"})]),
        # multiple flags (primary new feature)
        (
            ["--provides", "Pillow=PIL", "--provides", "PyJWT=jwt"],
            [("pillow", {"PIL"}), ("pyjwt", {"jwt"})],
        ),
        # normalization: uppercase → lowercase
        (["--provides", "PyJWT=jwt"], [("pyjwt", {"jwt"})]),
        # multiple modules provided by the same package
        (["--provides", "foo=fox,foo=bar"], [("foo", {"fox", "bar"})]),
        # normalization: hyphen → underscore
        (["--provides", "scikit-learn=sklearn"], [("scikit_learn", {"sklearn"})]),
        # normalization: mixed case + hyphen
        (["--provides", "SciKit-Learn=sklearn"], [("scikit_learn", {"sklearn"})]),
        # backward-compat: comma-separated in a single flag
        (
            ["--provides", "Pillow=PIL,PyJWT=jwt"],
            [("pillow", {"PIL"}), ("pyjwt", {"jwt"})],
        ),
    ],
)
def test__main__provides_parsing(
    monkeypatch: pytest.MonkeyPatch,
    argv_provides: list[str],
    expected_provides: list[tuple[str, set[str]]],
) -> None:
    """Test that --provides flags are parsed, merged, and normalized correctly."""
    captured: dict[str, Packages] = {}

    def _mock_yield(_file_names: object, app_cfg: AppConfig) -> object:
        captured["provides"] = app_cfg.provides
        return iter([])

    monkeypatch.setattr("check_dependencies.__main__.yield_wrong_imports", _mock_yield)
    monkeypatch.setattr(
        "sys.argv", ["check_dependencies", *argv_provides, DATA.as_posix()]
    )
    with pytest.raises(SystemExit):
        cli_main()
    packages = captured["provides"]
    for expected_pkg, expected_import in expected_provides:
        assert packages.modules(Package(expected_pkg)) == expected_import


class TestYieldWrongImports:
    """Test collection for the yield wrong imports function."""

    def fn(  # pylint: disable=too-many-arguments
        self,
        overwrite_cfg: Path = POETRY,
        file_names: Sequence[str] = (SRC,),
        include_dev: bool = False,
        verbose: bool = False,
        show_all: bool = False,
        known_extra: Sequence[str] = (),
        known_missing: Sequence[str] = (),
        provides: Sequence[str] = (),
    ) -> list[str]:
        """Call the yield wrong imports function with patched pyproject.toml."""
        with patch("check_dependencies.pyproject_toml._PYPROJECT_TOML", overwrite_cfg):
            return list(
                yield_wrong_imports(
                    file_names,
                    AppConfig.from_cli_args(
                        file_names=file_names,
                        include_dev=include_dev,
                        verbose=verbose,
                        show_all=show_all,
                        known_extra=",".join(known_extra),
                        known_missing=",".join(known_missing),
                        provides=provides or [],
                    ),
                ),
            )

    def test(self, pyproject: Path) -> None:
        """By default, we should only see the missing (and extra) imports."""
        assert self.fn(overwrite_cfg=pyproject) == [
            "! missing",
            "! missing_class",
            "! missing_def",
        ]

    def test_dev(self) -> None:
        """Test default with dev."""
        assert self.fn(overwrite_cfg=PYPROJECT_CFG, include_dev=True) == [
            "+ test_devtest > 0",
            "+ test_doctest > 0",
        ]

    def test_extra_requirements(self, pyproject_extra: Path) -> None:
        """Ensure extra requirements are printed by default."""
        res = self.fn(overwrite_cfg=pyproject_extra)
        spec = "= '> 0'" if pyproject_extra == POETRY_EXTRA else "> 0"
        assert res == [
            "! missing",
            "! missing_class",
            "! missing_def",
            f"+ test_extra {spec}",
        ]

    def test_extra_requirements_verbose(self, pyproject_extra: Path) -> None:
        """Ensure extra requirements are printed by default."""
        assert set(self.fn(overwrite_cfg=pyproject_extra, verbose=True)) > {
            "",
            "### Dependencies in config file not used in application:",
        }

    def test_extra_requirements_as_cfg(self) -> None:
        """Do not flog unused requirements passed in as an extra."""
        assert not self.fn(overwrite_cfg=PYPROJECT_CFG)

    def test_provides_from_config(self) -> None:
        """Packages with a provides mapping should not appear as missing or extra.

        The pyproject_pep631_provides.toml declares 'test_alias_pkg' as a dependency
        with ``provides.test_1 = "test_alias_pkg"``.  Since src.py imports test_1,
        neither ``! test_1`` (missing) nor ``+ test_alias_pkg`` (extra) should appear.
        """
        result = self.fn(overwrite_cfg=PYPROJECT_PROVIDES)
        assert "! test_1" not in result
        assert "+ test_alias_pkg" not in result
        assert result == ["! missing", "! missing_class", "! missing_def"]

    def test_provides_from_app_cfg(self) -> None:
        """AppConfig.provides (CLI --provides) resolves false positives like config."""
        # POETRY has test_alias_pkg NOT declared, but we pass provides via AppConfig
        # to map test_1 -> test_main (test_main IS declared in POETRY).
        result = self.fn(overwrite_cfg=POETRY, provides="test_main=test_1")
        assert "! test_1" not in result
        assert "+ test_main" not in result

    def test_provides_app_cfg_overrides_file_cfg(self) -> None:
        """CLI --provides takes precedence over [tool.check-dependencies.provides]."""
        # Config maps test_1 -> test_alias_pkg.  CLI overrides to test_1 -> test_main.
        result = self.fn(
            overwrite_cfg=PYPROJECT_PROVIDES,
            provides="test_main=test_1",
        )
        assert "! test_1" not in result
        assert "+ test_main" not in result

    def test_ignore_requirements(self, pyproject_extra: Path) -> None:
        """Ensure ignored requirements are not printed."""
        assert self.fn(overwrite_cfg=pyproject_extra, known_extra=["test_extra"]) == [
            "! missing",
            "! missing_class",
            "! missing_def",
        ]

    def test_ignore_requirements_still_check_in_src(self) -> None:
        """Ensure ignored requirements are not flagged even if they come up in src."""
        assert "  test_1" in self.fn(overwrite_cfg=POETRY, show_all=True)
        assert self.fn(overwrite_cfg=POETRY, known_missing=["test_1"]) == [
            "! missing",
            "! missing_class",
            "! missing_def",
        ]

    def test_show_all(self) -> None:
        """Show all imports, including correct ones."""
        assert self.fn(show_all=True) == [
            "! missing",
            "  test_1",
            "  test_main",
            "  check_dependencies",
            "! missing_class",
            "! missing_def",
        ]

    def test_include_extra(self) -> None:
        """Include development dependencies in the check."""
        res = self.fn(include_dev=True)
        assert [part.split("=", 1)[0].strip() for part in res] == [
            "! missing",
            "! missing_class",
            "! missing_def",
            "+ test_dev_1",
            "+ test_dev_2",
        ]

    def test_include_extra_requirements(self) -> None:
        """Test known_missing.

        Include requirements as known-missing - they should not appear in the output.
        """
        res = self.fn(known_missing=["missing", "test_1"])
        assert res == ["! missing_class", "! missing_def"]

    def test_verbose(self) -> None:
        """Verbose output should include the file and line number of the import."""
        assert self.fn(verbose=True) == [
            f"!NA {SRC}:1 missing.bar",
            f"!NA {SRC}:2 missing.foo",
            f"!NA {SRC}:5 missing",
            f"!NA {SRC}:11 missing_class",
            f"!NA {SRC}:15 missing",
            f"!NA {SRC}:16 missing_def",
        ]

    def test_verbose_show_all(self) -> None:
        """Test for the most verbose output."""
        assert self.fn(verbose=True, show_all=True) == [
            f"!NA {SRC}:1 missing.bar",
            f"!NA {SRC}:2 missing.foo",
            f" OK {SRC}:3 test_1",
            f" OK {SRC}:4 test_main",
            f"!NA {SRC}:5 missing",
            f" OK {SRC}:7 check_dependencies",
            f"!NA {SRC}:11 missing_class",
            f"!NA {SRC}:15 missing",
            f"!NA {SRC}:16 missing_def",
        ]

    @pytest.mark.parametrize("show_all", [True, False])
    @pytest.mark.parametrize("include_extra", [True, False])
    def test_directory_only_one_use(self, show_all: bool, include_extra: bool) -> None:
        """Print out only one instance of a missing import.

        Even for multiple files, make sure we only print out one instance of a
        missing import.
        """
        res = self.fn(
            file_names=[DATA.as_posix()], show_all=show_all, include_dev=include_extra
        )
        assert len(res) == len(set(res))

    def test_directory_both_files(self) -> None:
        """Given a directory, we should check all files for missing imports."""
        res = self.fn(file_names=[DATA.as_posix()])
        assert set(res) > {"! missing", "! missing_src2"}

    def test_all_imports_all_files(self) -> None:
        """show_all=True should show all imports in all files."""
        res = self.fn(file_names=[SRC_MODULE.as_posix()], show_all=True)
        assert set(res) == {
            "  check_dependencies",
            "  test_1",
            "  test_main",
            "! missing",
            "! missing_class",
            "! missing_def",
            "! missing_src2",
            "! tests_main",
        }

    def test_doublette_entries(self) -> None:
        """Test that doublette entries are not printed twice."""
        res = self.fn(file_names=[SRC_MODULE.as_posix()] * 2, show_all=True)
        assert sorted(res) == [
            "  check_dependencies",
            "  test_1",
            "  test_main",
            "! missing",
            "! missing_class",
            "! missing_def",
            "! missing_src2",
            "! tests_main",
        ]

    def test_no_fail_on_missing_source(self) -> None:
        """Test that we do not fail if the source file is missing."""
        res = AppConfig.from_cli_args(
            file_names=["nonexistent.py"],
            known_extra="",
            known_missing="",
            provides=[],
            include_dev=False,
            verbose=False,
            show_all=False,
        )
        assert res

    def test_unicode_imports(self) -> None:
        """Test for Unicode module names."""
        result = self.fn(overwrite_cfg=PYPROJECT_UNICODE, file_names=[SRC_UNICODE])

        # All Unicode module names should be flagged as missing
        assert result == []

    def test_unicode_imports_verbose(self) -> None:
        """Test verbose output with Unicode module names."""
        result = self.fn(file_names=[SRC_UNICODE], verbose=True)

        # Should show file path and line numbers for Unicode imports
        assert any("ö" in line and SRC_UNICODE in line for line in result)
        assert any("café" in line and SRC_UNICODE in line for line in result)

    def test_fail_msg_on_nonexisting_file(self) -> None:
        """Test non-existing files."""
        missing_file = (SRC_MODULE / "non-existing.pyy").as_posix()
        res = self.fn(file_names=[missing_file])
        assert res[0] == f"!! {missing_file}"


@pytest.mark.parametrize(
    "stmt, expected",
    [
        ("import foo", ["foo"]),
        ("import foo as bar", ["foo"]),
        ("from foo import bar", ["foo"]),
        ("from foo import bar as baz", ["foo"]),
        ("from foo import bar, baz", ["foo"]),
        ("from . import bar", []),
        ("from .internal import bar", []),
        ("import foo\nimport bar", ["foo", "bar"]),
        ("import foo.bar", ["foo.bar"]),
        ("class X:\n    import foo", ["foo"]),
        ("def x():\n    import foo", ["foo"]),
        ("try:\n    import foo\nexcept ImportError:\n    import bar", ["foo", "bar"]),
    ],
)
def test_imports_iter(stmt: str, expected: list[str]) -> None:
    """Test the imports iterator for statement junks."""
    parsed = ast.parse(dedent(stmt))
    assert [x[0] for x in _imports_iter(parsed.body)] == expected


def test_missing_import_iter_silent_on_invalid_python_code() -> None:
    """Test that missing imports iterator catches invalid Python code."""
    my_path = MagicMock()
    my_path.as_posix.return_value = "dummy.py"
    my_path.read_bytes.return_value = b"()foo"
    res = list(_missing_imports_iter(my_path, set(), Packages([])))
    assert len(res) == 1
    status, filename, _ = res[0]
    assert status == Dependency.FILE_ERROR
    assert filename == "dummy.py"


def test_missing_imports_iter_non_utf8_encoding(tmp_path: Path) -> None:
    """Test that _missing_imports_iter works with a non-UTF8-encoded file."""
    py_file = tmp_path / "latin1_module.py"
    # Write a latin-1 encoded file with an encoding cookie and an import
    content = "# -*- coding: latin-1 -*-\nimport os\nx = 'caf\xe9'\n"
    py_file.write_bytes(content.encode("latin-1"))
    result = list(_missing_imports_iter(py_file, set(), Packages([])))
    assert [m for _, m, _ in result] == ["os"]


def test_missing_imports_iter() -> None:
    """Test the missing import iterator."""
    res = list(
        _missing_imports_iter(
            Path(SRC), Package.set({"test_0", "test_1", "extra"}), Packages([])
        )
    )
    assert {c for c, _, _ in res} == {Dependency.NA, Dependency.OK}
    assert [m for _, m, _ in res] == [
        "missing.bar",
        "missing.foo",
        "test_1",
        "test_main",
        "missing",
        "check_dependencies",
        "missing_class",
        "missing",
        "missing_def",
    ]
    assert None not in {s for _, _, s in res}


@pytest.mark.parametrize(
    "verbose, expected",
    [
        (True, "+EXTRA foo"),
        (False, "+ foo"),
    ],
)
def test_mk_unused_formatter(verbose: bool, expected: str) -> None:
    """Test the unused formatter."""
    cfg = AppConfig.from_cli_args(file_names=[DATA.as_posix()], verbose=verbose)
    assert list(cfg.unused_fmt("foo")) == [expected]


@pytest.mark.parametrize(
    "stmt, expected",
    [
        # Unicode in module names (valid syntax, but won't work at runtime)
        ("import ö", ["ö"]),
        ("import café", ["café"]),
        ("from ä import something", ["ä"]),
        ("import 日本語", ["日本語"]),
        ("from Москва import test", ["Москва"]),
        # Unicode in variable names is valid, but shouldn't be confused with imports
        ("import os\nö = 1", ["os"]),
        # Mixed ASCII and Unicode
        ("import foo_ö", ["foo_ö"]),
    ],
)
def test_imports_iter_unicode(stmt: str, expected: list[str]) -> None:
    """Test that Unicode module names are handled correctly.

    While Python 3 allows Unicode identifiers (PEP 3131), module names
    must correspond to file names, which are typically ASCII. The tool
    should still parse these correctly, even though they won't work at runtime.
    """
    parsed = ast.parse(dedent(stmt))
    assert [x[0] for x in _imports_iter(parsed.body)] == expected


def test_missing_imports_iter_unicode_file(tmp_path: Path) -> None:
    """Test _missing_imports_iter with a file containing Unicode imports.

    This verifies that the tool can parse files with Unicode module names
    (even though such imports would fail at runtime).
    """
    py_file = tmp_path / "unicode_imports.py"
    content = """# -*- coding: utf-8 -*-
import ö
from café import something
import sys
"""
    py_file.write_text(content, encoding="utf-8")

    result = list(_missing_imports_iter(py_file, Package.set({"sys"}), Packages([])))

    # Extract module names
    modules = [m for _, m, _ in result]

    # Should have detected all three imports
    assert "ö" in modules
    assert "café" in modules
    assert "sys" in modules

    # sys should be OK, Unicode modules should be NA
    statuses = {m: status for status, m, _ in result}
    assert statuses["sys"] == Dependency.OK
    assert statuses["ö"] == Dependency.NA
    assert statuses["café"] == Dependency.NA


@pytest.mark.performance
def test_performance_large_project(tmp_path: Path) -> None:
    """Test the performance of yield_wrong_imports on a large project."""
    max_duration_per_file = 0.01  # in seconds
    n_files = 1000
    for i in range(n_files):
        (tmp_path / f"file_{i}.py").write_text("import sys\n")
    cfg = AppConfig(
        dependencies=[],
        known_extra=[],
        known_missing=[],
        provides=Packages([]),
    )
    start = time.time()
    _ = list(yield_wrong_imports([tmp_path.as_posix()], cfg))
    duration = time.time() - start

    assert duration < max_duration_per_file * n_files



"""Library for check_dependencies."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from enum import Enum
from functools import total_ordering
from itertools import groupby, takewhile
from operator import itemgetter
from typing import Iterable

logger = logging.getLogger("check_dependencies.lib")


class Dependency(Enum):
    """Possible dependency state."""

    NA = "!"  # Not Available
    EXTRA = "+"  # Extra dependency in config file
    OK = " "  # Correct import (declared in config file)
    FILE_ERROR = "!!"  # Error getting import statement (e.g. io error, syntax error)


def pkg(module: str) -> str:
    """Extract the top-level package name from a module import.

    **Examples:**
    >>> pkg("numpy.linalg")
    "numpy"
    >>> pkg("sklearn")
    "sklearn"
    >>> pkg("PIL.Image")
    "pil"

    :param module: Full module path (e.g., "package.submodule.module")
    :returns: Normalized top-level package name
    """
    return module.split(".", 1)[0].strip()


@dataclass(frozen=True)
@total_ordering
class Package:
    """A single package with its original and canonical name for comparison.

    The original name is preserved for display purposes, while the canonical name is
    used for hashing and comparison, ensuring that different representations of the
    same package are treated as equal.

    Added advantage is that we now have a clear distinction between a module
    (presented as a string) and a package.
    """

    __slots__ = ("_original", "canonical")
    _original: str
    canonical: str

    def __init__(self, package: str) -> None:
        """Initialize the Package dataclass."""
        object.__setattr__(self, "_original", package.strip())
        object.__setattr__(self, "canonical", _canonical(package))

    def __hash__(self) -> int:
        """Use only canonical name for hashing."""
        return hash(self.canonical)

    def __eq__(self, other: object) -> bool:
        """Compare with another package or a package name."""
        if isinstance(other, Package):
            return self.canonical == other.canonical
        if isinstance(other, str):
            return self.canonical == _canonical(other)
        return NotImplemented

    def __str__(self) -> str:
        """Get the Original name."""
        return self._original

    def __bool__(self) -> bool:
        """Check if there is a package (i.e. empty package name is Falsy)."""
        return bool(self.canonical)

    def __gt__(self, other: object) -> bool:
        """Compare with another package or a package name."""
        if isinstance(other, Package):
            return self.canonical > other.canonical
        if isinstance(other, str):
            return self.canonical > _canonical(other)
        return NotImplemented

    @classmethod
    def set(cls, package_names: Iterable[str]) -> __builtins__.set[Package]:
        """Get a set of packages from a package names."""
        return {cls(package_name) for package_name in package_names}


class Packages:
    """Translation layer to map between packages and modules."""

    _modules: dict[Package, set[str]]
    _packages: dict[str, set[Package]]

    def __init__(self, packages: list[tuple[Package, str]]) -> None:
        """Initialize the Packages dataclass.

        :param packages: List of (package, module) tuples, where package is the
            package name and module is the import name.
        """
        self._modules = {
            key: {module for _, module in val}
            for key, val in groupby(
                sorted(packages, key=itemgetter(0)), key=itemgetter(0)
            )
        }

        self._packages = {
            key: {pkg_ for pkg_, _ in val}
            for key, val in groupby(
                sorted(packages, key=itemgetter(1)), key=itemgetter(1)
            )
        }

    def modules(self, pkg_: str | Package) -> set[str]:
        """Get the modules (import name) for a given package name.

        :param pkg_: The package to look up.
        """
        if isinstance(pkg_, str):
            pkg_ = Package(pkg_)
        return self._modules.get(pkg_, {str(pkg_.canonical)})

    def packages(self, module_name: str) -> set[Package]:
        """Get the packages for a given module (import name).

        :param module_name: The module name (import name) to look up.
        """
        module_ = pkg(module_name)
        return self._packages.get(module_, {Package(module_)})


def _canonical(name: str) -> str:
    """Normalize a package name: lowercase and replace special chars with underscores.

    This makes package name comparison case-insensitive and treats hyphens and
    underscores as equivalent, consistent with PEP 503 / PyPI conventions.
    E.g. ``scikit-learn``, ``scikit_learn``, and ``SciKit-Learn`` all normalize
    to ``scikit_learn``.

    :param name: The package name to normalize.
    :returns: Normalized package name.
    """
    pkg_name_iter = takewhile(lambda x: x.isalnum() or x in "-_.", name.strip())
    package_name = "".join(pkg_name_iter)
    return package_name.lower().strip().replace("-", "_").replace(".", "_")



"""Parse project specific options (dependencies, config) from a pyproject.toml file."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from itertools import chain
from os.path import commonpath
from pathlib import Path
from typing import Any, Collection, Mapping, TypeVar

from check_dependencies.lib import Package

try:
    import tomllib  # type: ignore[unresolved-import]
except ImportError:  # pragma: no cover
    import toml as tomllib

logger = logging.getLogger(__name__)
_PYPROJECT_TOML = Path("pyproject.toml")
_T = TypeVar("_T")


@dataclass(frozen=True)
class PyProjectToml:
    """Project specific options (dependencies, config) from a pyproject.toml file."""

    cfg: dict[str, Any]
    path: Path
    include_dev: bool = False

    @classmethod
    def for_paths(
        cls, paths: Collection[str], *, include_dev: bool = False
    ) -> PyProjectToml:
        """Create a PyProjectToml instance from a pyproject.toml file.

        :param paths: List of file paths to analyze. The common parent directory
            will be searched for a pyproject.toml file.
        :param include_dev: Whether to include development dependencies
            from pyproject.toml.
        :returns: A PyProjectToml instance with the parsed configuration.
        """
        try:
            pyproject_candidate = Path(
                commonpath(Path(p).expanduser().resolve() for p in paths),
            )
        except ValueError as exc:
            # Can only be reached in Windows when two different drives are provided.
            msg = f"Error finding common path for {paths}: {exc}"
            raise ValueError(msg) from exc
        path = _get_pyproject_path(pyproject_candidate)

        return cls(
            cfg=tomllib.loads(path.read_text("utf-8")),
            path=path,
            include_dev=include_dev,
        )

    @property
    def dependencies(self) -> frozenset[Package]:
        """Get dependencies from pyproject.toml file."""
        deps: set[Package] = set()
        is_used = False
        for dep_class in (
            _Pep621Dependencies,
            _PoetryDependencies,
            _UvLegacyDependencies,
            _HatchDependencies,
        ):
            dep_cfg = dep_class(self.cfg)
            if dep_cfg.is_used():
                is_used = True
                deps |= dep_cfg.dependencies(include_dev=self.include_dev)
        if not is_used:
            msg = "No dependency management found"
            raise ValueError(msg)
        return frozenset(deps)

    @property
    def known_missing(self) -> frozenset[str]:
        """Known to be used in application but not declared in requirements.

        This includes the project itself.
        """
        # Add project name
        pep631_name = Package(_nested_item(self.cfg, "project.name", str) or "")
        poetry_name = Package(_nested_item(self.cfg, "tool.poetry.name", str) or "")
        return frozenset(
            filter(
                None,
                (
                    pep631_name.canonical,
                    poetry_name.canonical,
                    *_nested_item(
                        self.cfg,
                        "tool.check-dependencies.known-missing",
                        list,
                    ),
                ),
            ),
        )

    @property
    def known_extra(self) -> frozenset[Package]:
        """Dependencies that are known to be unused in application."""
        return frozenset(
            map(
                Package,
                _nested_item(self.cfg, "tool.check-dependencies.known-extra", list),
            )
        )

    @property
    def provides(self) -> list[tuple[Package, str]]:
        """Mapping from import name to package name.

        E.g. ``{"jwt": "pyjwt", "shapefile": "pyshp"}`` means that the package
        ``pyjwt`` is imported as ``jwt`` and ``pyshp`` as ``shapefile``.

        Package keys are canonicalized via :class:`Package` (case-insensitive,
        hyphen/underscore equivalent), so e.g. ``PyJWT``, ``pyjwt``, and
        ``py-jwt`` resolve to the same package identity.
        """
        return [
            (Package(package), module)
            for package, modules in _nested_item(
                self.cfg, "tool.check-dependencies.provides", dict
            ).items()
            for module in ([modules] if isinstance(modules, str) else modules)
        ]


@dataclass(frozen=True)
class _BaseDependency:
    """Base class for different dependency providers."""

    cfg: Mapping[str, Any]

    def is_used(self) -> bool:
        """Check if the pyproject.toml file contains this style of dependencies."""
        raise NotImplementedError  # pragma: no cover

    def dependencies(self, *, include_dev: bool) -> set[Package]:
        """Get all dependencies.

        :arg include_dev: Whether to include dev dependencies.
        """
        deps = self._dependencies()
        if include_dev:
            deps.update(self._dev_dependencies())
        return deps

    def _dev_dependencies(self) -> set[Package]:
        """Get development dependencies."""
        raise NotImplementedError  # pragma: no cover

    def _dependencies(self) -> set[Package]:
        """Get production dependencies, with extras, but no development dependencies."""
        raise NotImplementedError  # pragma: no cover


@dataclass(frozen=True)
class _Pep621Dependencies(_BaseDependency):
    """PEP-621 dependency provider."""

    def is_used(self) -> bool:
        """Check if the pyproject.toml file contains PEP 621-style dependencies."""
        return "dependencies" in self.cfg.get("project", {})

    def _dependencies(self) -> set[Package]:
        """Get dependencies from a PEP 621-style pyproject.toml file."""
        deps = Package.set(_nested_item(self.cfg, "project.dependencies", list))

        for raw_extras in _nested_item(
            self.cfg, "project.optional-dependencies", dict
        ).values():
            deps.update(Package.set(raw_extras))
        return deps

    def _dev_dependencies(self) -> set[Package]:
        """Get the dev dependencies from a PEP 621-style pyproject.toml file."""
        groups = _nested_item(self.cfg, "dependency-groups", dict)
        return set().union(*map(Package.set, groups.values()))


@dataclass(frozen=True)
class _PoetryDependencies(_BaseDependency):
    """Poetry Dependencies."""

    def is_used(self) -> bool:
        """Check if the pyproject.toml file contains Poetry style dependencies."""
        return bool(_nested_item(self.cfg, "tool.poetry", dict))

    def _dependencies(self) -> set[Package]:
        poetry_deps = dict(_nested_item(self.cfg, "tool.poetry.dependencies", dict))
        poetry_deps.pop("python", None)
        return Package.set(self._names_from_items(poetry_deps))

    def _dev_dependencies(self) -> set[Package]:
        # Get tool.poetry.group.*.dependencies
        # e.g. groups is "dev": {"dependencies": {"pytest": "^6.2.5"}}
        deps = Package.set(
            self._names_from_items(
                _nested_item(self.cfg, "tool.poetry.dev-dependencies", dict)
            )
        )
        groups: dict[str, dict[str, dict[str, str]]] = _nested_item(
            self.cfg, "tool.poetry.group", dict
        )
        for group in groups.values():
            deps |= Package.set(self._names_from_items(group.get("dependencies", {})))
        return deps

    @staticmethod
    def _names_from_items(items: dict[str, Any]) -> list[str]:
        """Get the package name from a Poetry dependency item."""
        return [f"{k} = {v!r}" for k, v in items.items()]


@dataclass(frozen=True)
class _UvLegacyDependencies(_BaseDependency):
    """uv (legacy) dependency manager."""

    def is_used(self) -> bool:
        """Check if uv is used."""
        return bool(_nested_item(self.cfg, "tool.uv", dict))

    def _dependencies(self) -> set[Package]:
        return _Pep621Dependencies(self.cfg).dependencies(include_dev=False)

    def _dev_dependencies(self) -> set[Package]:
        return Package.set(_nested_item(self.cfg, "tool.uv.dev-dependencies", dict))


@dataclass(frozen=True)
class _HatchDependencies(_BaseDependency):
    """Hatch Dependencies."""

    def is_used(self) -> bool:
        """Check if hatch is used in this project."""
        return bool(_nested_item(self.cfg, "tool.hatch", dict))

    def _dependencies(self) -> set[Package]:
        return Package.set(
            _nested_item(self.cfg, "tool.hatch.envs.default.dependencies", list)
        )

    def _dev_dependencies(self) -> set[Package]:
        return set().union(
            *(
                Package.set(env_cfg.get("dependencies", []))
                for name, env_cfg in _nested_item(
                    self.cfg, "tool.hatch.envs", dict
                ).items()
                if name != "default"
            )
        )


def _nested_item(obj: Mapping[str, Any], key: str, /, class_: type[_T]) -> _T:
    """Get items from a nested dictionary where the keys are dot-separated.

    :param key: The dot-separated key to look up in the nested dictionary.
    :param class_: The expected type of the value.
    :returns: The value corresponding to the key if found
        otherwise the default instance of the expected type.
    :raises TypeError: If the value found is not of the expected type or if any part of
        key is not a mapping.
    """
    for a in key.split("."):
        if a not in obj:
            return class_()
        obj = obj[a]
    if not isinstance(obj, class_):
        msg = f"Expected {class_} but got {type(obj)}"
        raise TypeError(msg)
    return obj


def _get_pyproject_path(path: Path) -> Path:
    """Get the pyproject.toml file by searching up the directory hierarchy.

    :param path: The starting path to search from.
    """
    for p in chain([path], path.resolve().parents):
        if (p / _PYPROJECT_TOML).exists():
            return p / _PYPROJECT_TOML
    msg = f"Could not find {_PYPROJECT_TOML} file within path hierarchy"
    raise FileNotFoundError(msg)



# Test file with Unicode imports
# These should fail or be handled appropriately

import ö
from ä import something
import café
from naïve import something_else
import 日本語
from Москва import test
from москва2 import test2




"""Tests for the builtin modules."""

from __future__ import annotations

from check_dependencies.builtin_module import BUILTINS


def test_is_set() -> None:
    """Are the builtin modules stored in a set?"""
    assert isinstance(BUILTINS, frozenset)


def test_contains_future() -> None:
    """Test a single sample module."""
    assert "__future__" in BUILTINS


def test_no_empty_module() -> None:
    """Ensure no empty string is in the builtins set."""
    assert "" not in BUILTINS


def test_all_correct_names() -> None:
    """Ensure all builtin modules are valid identifiers."""
    for module in BUILTINS:
        assert module.isidentifier()
